"""
Unit tests for the git module.
"""
import collections
import subprocess
from pytest_mock import MockerFixture
import instarepo.git


def test_clone(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")

    # act
    result = instarepo.git.clone("ssh://hello.git", "/tmp/hello")

    # assert
    mock.assert_called_once_with(
        ["git", "clone", "ssh://hello.git", "/tmp/hello"], check=True
    )
    assert result.dir == "/tmp/hello"


def test_clone_quietly(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")

    # act
    result = instarepo.git.clone("ssh://hello.git", "/tmp/hello", quiet=True)

    # assert
    mock.assert_called_once_with(
        ["git", "clone", "-q", "ssh://hello.git", "/tmp/hello"], check=True
    )
    assert result.dir == "/tmp/hello"


def test_create_branch(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    git.create_branch("release")

    # assert
    mock.assert_called_once_with(
        ["git", "checkout", "-b", "release"], check=True, cwd="/tmp/hello"
    )


def test_create_branch_quietly(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello", quiet=True)

    # act
    git.create_branch("release")

    # assert
    mock.assert_called_once_with(
        ["git", "checkout", "-q", "-b", "release"], check=True, cwd="/tmp/hello"
    )


def test_add(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    git.add("test.txt")

    # assert
    mock.assert_called_once_with(
        ["git", "add", "test.txt"], check=True, cwd="/tmp/hello"
    )


def test_commit(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    git.commit("oops")

    # assert
    mock.assert_called_once_with(
        ["git", "commit", "-m", "oops"], check=True, cwd="/tmp/hello"
    )


def test_commit_quietly(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello", quiet=True)

    # act
    git.commit("oops")

    # assert
    mock.assert_called_once_with(
        ["git", "commit", "-q", "-m", "oops"], check=True, cwd="/tmp/hello"
    )


def test_push(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    git.push()

    # assert
    mock.assert_called_once_with(
        ["git", "push", "-u", "origin", "HEAD"],
        check=True,
        cwd="/tmp/hello",
    )


def test_delete_remote_branch(mocker: MockerFixture):
    # arrange
    mock = mocker.patch("subprocess.run")
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    git.delete_remote_branch("obsolete-branch")

    # assert
    mock.assert_called_once_with(
        ["git", "push", "--delete", "origin", "obsolete-branch"],
        check=True,
        cwd="/tmp/hello",
    )


def test_join():
    git = instarepo.git.GitWorkingDir("/tmp/hello")
    filename = git.join("src", "index.js").replace("\\", "/")
    assert filename == "/tmp/hello/src/index.js"


def test_join_with_slashes_in_string():
    git = instarepo.git.GitWorkingDir("/tmp/hello")
    filename = git.join("src/index.js").replace("\\", "/")
    assert filename == "/tmp/hello/src/index.js"


def test_is_remote_branch_present(mocker: MockerFixture):
    # arrange
    result_type = collections.namedtuple("ProcessResult", ["stdout"])
    result = result_type("remote-ref")
    mock = mocker.patch("subprocess.run")
    mock.return_value = result
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    result = git.is_remote_branch_present("existing-branch")

    # assert
    mock.assert_called_once_with(
        ["git", "rev-parse", "-q", "--verify", "remotes/origin/existing-branch"],
        check=True,
        cwd="/tmp/hello",
        encoding="utf-8",
        stdout=subprocess.PIPE,
    )
    assert result


def test_is_remote_branch_absent(mocker: MockerFixture):
    # arrange
    result_type = collections.namedtuple("ProcessResult", ["stdout"])
    result = result_type("")
    mock = mocker.patch("subprocess.run")
    mock.return_value = result
    git = instarepo.git.GitWorkingDir("/tmp/hello")

    # act
    result = git.is_remote_branch_present("missing-branch")

    # assert
    mock.assert_called_once_with(
        ["git", "rev-parse", "-q", "--verify", "remotes/origin/missing-branch"],
        check=True,
        cwd="/tmp/hello",
        encoding="utf-8",
        stdout=subprocess.PIPE,
    )
    assert not result



"""Fixes for Maven projects"""
import logging
import os
import os.path
import platform
import re
import subprocess
import tempfile
import xml.etree.ElementTree as ET
from typing import Optional

import requests

import instarepo.git
import instarepo.github
import instarepo.xml_utils
from instarepo.fixers.base import MissingFileFix
from .finders import is_maven_project
from .readme import locate_badges, merge_badges


LOG_LEVEL = re.compile(r"^\[[A-Z]+\]")


def filter_maven_output(output: str) -> str:
    lines = output.splitlines()
    modified_lines = (strip_log_level(line) for line in lines)
    filtered_lines = (line for line in modified_lines if filter_line(line))
    return os.linesep.join(filtered_lines)


def strip_log_level(line: str) -> str:
    line = LOG_LEVEL.sub("", line)
    line = line.strip()
    return line


def filter_line(line: str) -> bool:
    if not line:
        return False
    allow_prefixes = ["Updated ", "Updating "]
    for allow_prefix in allow_prefixes:
        if line.startswith(allow_prefix):
            return True
    return False


def get_latest_artifact_version(group_id: str, artifact_id: str) -> Optional[str]:
    """
    Gets the latest published artifact version from central maven.

    Example url: https://repo1.maven.org/maven2/com/github/ngeor/archetype-quickstart-jdk8/maven-metadata.xml

    File structure:

    ```xml
    <?xml version="1.0" encoding="UTF-8"?>
    <metadata>
        <groupId>com.github.ngeor</groupId>
        <artifactId>archetype-quickstart-jdk8</artifactId>
        <versioning>
            <latest>2.8.0</latest>
            <release>2.8.0</release>
            <versions>
                <version>1.0.14</version>
                <version>1.0.22</version>
                <version>1.0.27</version>
                <version>1.0.29</version>
                <version>1.1.0</version>
                <version>1.1.1</version>
                <version>1.1.2</version>
                <version>1.2.0</version>
                <version>1.3.0</version>
                <version>1.4.0</version>
                <version>2.0.0</version>
                <version>2.1.0</version>
                <version>2.2.0</version>
                <version>2.3.0</version>
                <version>2.4.0</version>
                <version>2.5.0</version>
                <version>2.8.0</version>
            </versions>
            <lastUpdated>20210925070507</lastUpdated>
        </versioning>
    </metadata>
    ```
    """
    group_path = group_id.replace(".", "/")
    url = (
        f"https://repo1.maven.org/maven2/{group_path}/{artifact_id}/maven-metadata.xml"
    )
    response = requests.get(url)
    response.raise_for_status()
    root = ET.fromstring(response.text)
    versioning = root.find("versioning")
    if versioning is not None:
        release = versioning.find("release")
        if release is not None:
            return release.text

    logging.warning("URL %s returned unexpected XML", url)
    return None


MAVEN_YML = """# This workflow will build a Java project with Maven, and cache/restore any dependencies to improve the workflow execution time
# For more information see: https://help.github.com/actions/language-and-framework-guides/building-and-testing-java-with-maven

name: Java CI with Maven

on:
  push:
    branches: [ trunk ]
  pull_request:
    branches: [ trunk ]

jobs:
  build:

    runs-on: ubuntu-latest

    steps:
    - uses: actions/checkout@v2
    - name: Set up JDK 11
      uses: actions/setup-java@v2
      with:
        java-version: '11'
        distribution: 'adopt'
        cache: maven
    - name: Build with Maven
      run: mvn -B package --file pom.xml
"""


class MustHaveMavenGitHubWorkflowFix(MissingFileFix):
    """If missing, adds a GitHub action Maven build workflow"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        super().__init__(git, ".github/workflows/maven.yml")

    def should_process_repo(self):
        return is_maven_project(self.git.dir)

    def get_contents(self):
        return MAVEN_YML


class MavenBadgesFix:
    """
    Fixes badges for Maven libraries.

    Does not work for local git repositories.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        repo: Optional[instarepo.github.Repo],
        **kwargs,
    ):
        self.git = git
        self.repo = repo
        self.maven = Maven(git.dir)

    def run(self):
        if not self.git.isfile("README.md"):
            return []
        if not self.git.isfile("pom.xml"):
            return []
        if not self.repo:
            return []
        badges = self._badges_dict(self.repo)

        if not badges:
            return []

        with open(self.git.join("README.md"), "r", encoding="utf-8") as file:
            before_badges, existing_badges, after_badges = locate_badges(file.read())
            has_changes = False
            for i in range(len(existing_badges)):
                existing_badge = existing_badges[i]
                for needle, markdown in list(badges.items()):
                    if needle in existing_badge:
                        # delete the badge from the dictionary
                        # so that we don't add it again later
                        del badges[needle]
                        if existing_badge != markdown:
                            has_changes = True
                            existing_badges[i] = markdown
                        break
            if badges:
                existing_badges.extend(badges.values())
                has_changes = True
        if not has_changes:
            return []
        with open(self.git.join("README.md"), "w", encoding="utf-8") as file:
            file.write(merge_badges(before_badges, existing_badges, after_badges))
        self.git.add("README.md")
        msg = "Updated Maven badges in README.md"
        self.git.commit(msg)
        return [msg]

    def _badges_dict(self, repo: instarepo.github.Repo):
        badges = {}
        badges.update(self._github_actions_badge(repo))
        badges.update(self._effective_pom_badges())
        return badges

    def _github_actions_badge(self, repo: instarepo.github.Repo):
        if self.git.isfile(".github", "workflows", "maven.yml"):
            needle = "actions/workflows"
            markdown = f"[![Java CI with Maven](https://github.com/{repo.full_name}/actions/workflows/maven.yml/badge.svg)](https://github.com/{repo.full_name}/actions/workflows/maven.yml)"
            return {needle: markdown}
        return {}

    def _effective_pom_badges(self):
        badges = {}
        with tempfile.TemporaryDirectory() as tmp_dir:
            filename = os.path.join(tmp_dir, "pom.xml")
            self.maven.run("-B", "help:effective-pom", f"-Doutput={filename}")
            tree = instarepo.xml_utils.parse(filename)
            root = tree.getroot()
            badges.update(maven_central_badge(root))
            # edge case for checkstyle-rules artifact which has to publish
            # javadoc but doesn't have any source code
            if self.git.isdir("src", "main", "java"):
                badges.update(javadoc_badge(root))
        return badges


def maven_central_badge(root: ET.Element):
    group_id = root.findtext("{http://maven.apache.org/POM/4.0.0}groupId")
    if not group_id:
        return {}
    artifact_id = root.findtext("{http://maven.apache.org/POM/4.0.0}artifactId")
    if not artifact_id:
        return {}
    group_id_as_path = group_id.replace(".", "/")
    url = f"https://repo1.maven.org/maven2/{group_id_as_path}/{artifact_id}/"
    response = requests.get(url)
    if not response.ok:
        return {}
    needle = "maven-central"
    markdown = f"[![Maven Central](https://img.shields.io/maven-central/v/{group_id}/{artifact_id}.svg?label=Maven%20Central)](https://search.maven.org/search?q=g:%22{group_id}%22%20AND%20a:%22{artifact_id}%22)"
    return {needle: markdown}


def javadoc_badge(root: ET.Element):
    plugins = instarepo.xml_utils.find(
        root,
        "{http://maven.apache.org/POM/4.0.0}build",
        "{http://maven.apache.org/POM/4.0.0}plugins",
    )
    if plugins is None:
        return {}
    uses_maven_source_plugin = False
    uses_maven_javadoc_plugin = False
    for plugin in plugins:
        plugin_artifact_id = plugin.findtext(
            "{http://maven.apache.org/POM/4.0.0}artifactId"
        )
        if plugin_artifact_id == "maven-source-plugin":
            if "jar-no-fork" in plugin_goals(plugin):
                uses_maven_source_plugin = True
        if plugin_artifact_id == "maven-javadoc-plugin":
            if "jar" in plugin_goals(plugin):
                uses_maven_javadoc_plugin = True
    if not (uses_maven_source_plugin and uses_maven_javadoc_plugin):
        return {}
    group_id = root.findtext("{http://maven.apache.org/POM/4.0.0}groupId")
    if not group_id:
        return {}
    artifact_id = root.findtext("{http://maven.apache.org/POM/4.0.0}artifactId")
    if not artifact_id:
        return {}
    needle = "javadoc"
    markdown = f"[![javadoc](https://javadoc.io/badge2/{group_id}/{artifact_id}/javadoc.svg)](https://javadoc.io/doc/{group_id}/{artifact_id})"
    return {needle: markdown}


def plugin_goals(plugin: ET.Element):
    executions = plugin.find("{http://maven.apache.org/POM/4.0.0}executions")
    if executions is None:
        return set()
    executions = executions.findall("{http://maven.apache.org/POM/4.0.0}execution")
    goals = map(
        lambda x: x.find("{http://maven.apache.org/POM/4.0.0}goals"),
        executions,
    )
    goals = [x for x in goals if x]
    return set(
        y.text
        for x in goals
        for y in x.findall("{http://maven.apache.org/POM/4.0.0}goal")
        if y.text
    )


class Maven:
    """
    Runs Maven commands.
    """

    def __init__(self, directory: str):
        """
        Creates an instance of this class.

        :param directory: The directory of a Maven project
        """
        self.directory = directory

    def run(self, *args) -> str:
        """
        Runs Maven commands.
        """
        mvn_executable = "mvn.cmd" if platform.system() == "Windows" else "mvn"
        result = subprocess.run(
            [mvn_executable, *args],
            cwd=self.directory,
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        maven_output = result.stdout.strip()
        if result.returncode != 0:
            raise ChildProcessError(maven_output)
        return maven_output

    def sort_pom(self):
        self.run(
            "-B",
            "com.github.ekryd.sortpom:sortpom-maven-plugin:sort",
            "-Dsort.createBackupFile=false",
        )


class UrlFix:
    """
    Ensures Maven projects have the correct URL and SCM sections.

    Does not work for local git repositories.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        repo: Optional[instarepo.github.Repo],
        **kwargs,
    ):
        self.git = git
        self.repo = repo

    def run(self):
        if not self.git.isfile("pom.xml") or not self.repo:
            return []
        try:
            ET.register_namespace("", "http://maven.apache.org/POM/4.0.0")
            return self.do_run(self.repo)
        finally:
            ET.register_namespace("maven", "http://maven.apache.org/POM/4.0.0")

    def do_run(self, repo: instarepo.github.Repo):
        has_changes = False
        tree = instarepo.xml_utils.parse(self.git.join("pom.xml"))
        root = tree.getroot()
        has_changes |= ensure_element(
            root, "{http://maven.apache.org/POM/4.0.0}url", repo.html_url
        )
        scm = root.find("{http://maven.apache.org/POM/4.0.0}scm")
        if scm is None:
            has_changes = True
            scm = ET.Element("{http://maven.apache.org/POM/4.0.0}scm")
            root.append(scm)
        has_changes |= ensure_element(
            scm,
            "{http://maven.apache.org/POM/4.0.0}connection",
            f"scm:git:{repo.clone_url}",
        )
        has_changes |= ensure_element(
            scm,
            "{http://maven.apache.org/POM/4.0.0}developerConnection",
            f"scm:git:{repo.ssh_url}",
        )
        has_changes |= ensure_element(
            scm, "{http://maven.apache.org/POM/4.0.0}tag", "HEAD"
        )
        has_changes |= ensure_element(
            scm, "{http://maven.apache.org/POM/4.0.0}url", repo.html_url
        )
        if has_changes:
            tree.write(self.git.join("pom.xml"))
            Maven(self.git.dir).sort_pom()
            self.git.add("pom.xml")
            msg = "Corrected url info in pom.xml"
            self.git.commit(msg)
            return [msg]

        return []


def ensure_element(parent: ET.Element, child_name: str, child_text: str) -> bool:
    has_changes = False
    child_element = parent.find(child_name)
    if child_element is None:
        has_changes = True
        child_element = ET.Element(child_name)
        parent.append(child_element)
    if child_element.text != child_text:
        has_changes = True
        child_element.text = child_text
    return has_changes



"""Unit tests for readme.py"""

import re
from .readme import locate_badges, merge_badges, RE_MARKDOWN_IMAGE


def test_match_screenshot():
    """Tests replacing a screenshot with the module's regex"""
    input_readme = """
    # GodFather
A Delphi app to rename files (legacy project)

![screenshot](/GodFather/scrnshot.png?raw=true "Screenshot")
    """
    expected = """
    # GodFather
A Delphi app to rename files (legacy project)

![screenshot](/scrnshot.png?raw=true "Screenshot")
    """

    def replacer(match: re.Match) -> str:
        """The replacer function for the regex replacement"""
        return (
            match.string[match.start() : match.start("filename")]
            + "/scrnshot.png"
            + match.string[match.end("filename") : match.end()]
        )

    actual = RE_MARKDOWN_IMAGE.sub(replacer, input_readme)
    assert actual == expected


def test_locate_badges_one_badge():
    """Tests locating badges with a single badge"""
    input_readme = """# archetype-quickstart-jdk8

[![Maven Central](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8/badge.svg)](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8)

A Maven archetype for a simple Java app, updated for Java 8.

This is effectively the same as the maven-archetype-quickstart,
"""

    expected = (
        """# archetype-quickstart-jdk8
""",
        [
            "[![Maven Central](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8/badge.svg)](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8)"
        ],
        """A Maven archetype for a simple Java app, updated for Java 8.

This is effectively the same as the maven-archetype-quickstart,
""",
    )
    actual = locate_badges(input_readme)
    assert expected == actual


def test_locate_badges_no_badges():
    """Tests locating badges when no badges exist"""
    input_readme = """# some project

Some project description that goes on for a while
and even wraps multiple lines.

Some other text second paragraph here.
"""
    expected = (
        """# some project
""",
        [],
        """Some project description that goes on for a while
and even wraps multiple lines.

Some other text second paragraph here.
""",
    )
    actual = locate_badges(input_readme)
    assert expected == actual


def test_locate_and_merge_badges_one_badge():
    """Tests locating badges with a single badge and merging it back"""
    input_readme = """# archetype-quickstart-jdk8

[![Maven Central](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8/badge.svg)](https://maven-badges.herokuapp.com/maven-central/com.github.ngeor/archetype-quickstart-jdk8)

A Maven archetype for a simple Java app, updated for Java 8.

This is effectively the same as the maven-archetype-quickstart,
"""
    before, badges, after = locate_badges(input_readme)
    merged = merge_badges(before, badges, after)
    assert input_readme == merged


def test_locate_and_merge_badges_no_badges():
    """Tests locating and merging badges when no badges exist"""
    input_readme = """# some project

Some project description that goes on for a while
and even wraps multiple lines.

Some other text second paragraph here.
"""
    before, badges, after = locate_badges(input_readme)
    merged = merge_badges(before, badges, after)
    assert input_readme == merged



"""Unit tests for dotnet.py"""

from .dotnet import get_projects_from_sln_file_contents


def test_get_projects_from_sln_file_contents():
    """Tests extracting csproj files from sln contents"""
    sln_file = """# Visual Studio 15
VisualStudioVersion = 15.0.27130.2020
MinimumVisualStudioVersion = 10.0.40219.1
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender.Web", "CVRender.Web\\CVRender.Web.csproj", "{46D05687-EB9B-4885-9A14-1BDC8BBB253B}"
EndProject
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender", "CVRender\\CVRender.csproj", "{BD17C766-DF9E-4117-A8CB-2BAA8FE6D9B9}"
EndProject
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender.Tests", "CVRender.Tests\\CVRender.Tests.csproj", "{FDABDD4B-8BF5-4E4A-B977-400D0CE04D4A}"
EndProject
Project("{2150E333-8FDC-42A3-9474-1A3956D46DE8}") = "Solution Items", "Solution Items", "{184005A4-82D4-489E-BD9C-390DFEBC074D}"
	ProjectSection(SolutionItems) = preProject
		.gitignore = .gitignore
		appveyor.yml = appveyor.yml
		LICENSE = LICENSE
		README.md = README.md
	EndProjectSection
EndProject"""
    projects = list(get_projects_from_sln_file_contents(sln_file))
    assert projects == [
        "CVRender.Web\\CVRender.Web.csproj",
        "CVRender\\CVRender.csproj",
        "CVRender.Tests\\CVRender.Tests.csproj",
    ]



import os
import os.path
from typing import List
import instarepo.git


class CompositeFix:
    def __init__(self, rules):
        self.rules = rules

    def run(self):
        result: List[str] = []
        for rule in self.rules:
            result.extend(rule.run())
        return result


class SingleFileFix:
    def __init__(self, git: instarepo.git.GitWorkingDir, filename: str, msg: str):
        self.git = git
        self.filename = filename
        self.msg = msg

    def run(self):
        filename = os.path.join(self.git.dir, self.filename)
        if not os.path.isfile(filename):
            return []
        with open(filename, "r", encoding="utf-8") as file:
            contents = file.read()
        converted_contents = self.convert(contents)
        if contents == converted_contents:
            return []
        with open(filename, "w", encoding="utf-8") as file:
            file.write(converted_contents)
        self.git.add(self.filename)
        self.git.commit(self.msg)
        return [self.msg]

    def convert(self, contents: str) -> str:
        return contents


class MissingFileFix:
    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        filename: str,
    ):
        self.git = git
        if not filename:
            raise ValueError("filename cannot be empty")
        parts = filename.replace("\\", "/").split("/")
        for part in parts:
            if not part:
                raise ValueError(f"Found empty path segment in {filename}")
        self.directory_parts = parts[0:-1]
        self.filename_part = parts[-1]

    def run(self):
        self._ensure_directories()
        relative_filename = os.path.join(*self.directory_parts, self.filename_part)
        full_filename = os.path.join(self.git.dir, relative_filename)
        if os.path.isfile(full_filename):
            return []
        if not self.should_process_repo():
            return []
        contents = self.get_contents()
        if not contents:
            return []
        with open(full_filename, "w", encoding="utf8") as file:
            file.write(contents)
        self.git.add(relative_filename)
        msg = "chore: Adding " + relative_filename
        self.git.commit(msg)
        return [msg]

    def get_contents(self) -> str:
        return ""

    def should_process_repo(self) -> bool:
        return True

    def _ensure_directories(self):
        ensure_directories(self.git, *self.directory_parts)


def ensure_directories(git: instarepo.git.GitWorkingDir, *args):
    """
    Ensures that the given directories exist in the Git working directory.
    You can provide directories one by one, or separated by slash.
    """
    root = git.dir
    for directory in args:
        parts = directory.replace("\\", "/").split("/")
        for part in parts:
            root = os.path.join(root, part)
            if not os.path.isdir(root):
                os.mkdir(root)



"""Unit tests for finders.py"""

from .finders import is_file_of_extension


class TestIsFileOfExtension:
    """Unit tests for is_file_of_extension"""

    def test_directory(self):
        entry = MockDirEntry("hello.vbp", is_file=False)
        assert not is_file_of_extension(entry, ".vbp")
        assert not is_file_of_extension(entry, ".sln")
        assert not is_file_of_extension(entry)

    def test_file(self):
        entry = MockDirEntry("hello.vbp")
        assert is_file_of_extension(entry)
        assert is_file_of_extension(entry, ".vbp")
        assert not is_file_of_extension(entry, ".sln")
        assert is_file_of_extension(entry, ".vbg", ".vbp")
        assert not is_file_of_extension(entry, ".sln", ".xml")
        assert not is_file_of_extension(entry, ".bp")


class MockDirEntry:
    """Mock directory entry"""

    def __init__(self, name, is_file=True):
        """Creates an instance of this class"""
        self.name = name
        self._is_file = is_file

    def is_file(self):
        """Returns True if the entry is a file"""
        return self._is_file



import logging
import os
import tempfile
from datetime import date, datetime

import instarepo.git
import instarepo.github
import instarepo.repo_source


class AnalyzeCommand:
    def __init__(self, args):
        self.repo_source = (
            instarepo.repo_source.RepoSourceBuilder().with_args(args).build()
        )
        self.verbose: bool = args.verbose
        # TODO find date of oldest commit:
        # $ git show -s --format=%ci $(git rev-list --max-parents=0 HEAD) | cut -d' ' -f1
        # 2017-02-18
        self.since = datetime.strptime(args.since, "%Y-%m-%d").date()
        self.metric: str = args.metric

    def run(self):
        repos = self.repo_source.get()
        for repo in repos:
            self.process(repo)

    def process(self, repo: instarepo.github.Repo):
        logging.info("Processing repo %s", repo.name)
        with tempfile.TemporaryDirectory() as tmpdirname:
            logging.debug("Cloning repo into temp dir %s", tmpdirname)
            git = instarepo.git.clone(repo.ssh_url, tmpdirname, quiet=not self.verbose)
            self.process_git(git)

    def process_git(self, git: instarepo.git.GitWorkingDir):
        if self.metric == "commits":
            self.print_commits(git)
        elif self.metric == "files":
            self.print_files(git)
        else:
            raise ValueError("Unknown metric " + self.metric)

    def print_commits(self, git: instarepo.git.GitWorkingDir):
        print("Since,Until,Commit Count")
        for dt in generate_date_range(self.since):
            since, until = dt, next_month(dt)
            result = git.rev_list(since, until)
            print(f"{since},{until},{len(result)}")

    def print_files(self, git: instarepo.git.GitWorkingDir):
        print("Since,Until,Commit Date,File Count,LOC")
        current_branch_name = git.current_branch_name()
        for dt in generate_date_range(self.since):
            since, until = dt, next_month(dt)
            oldest_commit_id = oldest_commit_between_dates(git, since, until)
            if oldest_commit_id:
                git.checkout(oldest_commit_id)
                oldest_commit_date = git.commit_date(oldest_commit_id)

                stats = {"file": 0, "loc": 0}
                stats = count_files(git.dir, composite_counter, stats)
                print(
                    f'{since},{until},{oldest_commit_date},{stats["file"]},{stats["loc"]}'
                )
                git.checkout(current_branch_name)


def generate_date_range(since):
    stop = date.today()
    i = since
    while i < stop:
        yield i
        i = next_month(i)


def next_month(dt):
    if dt.month == 12:
        return date(dt.year + 1, 1, 1)
    else:
        return date(dt.year, dt.month + 1, 1)


def count_files(path, visitor, result):
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                result = visitor(result, path, entry.name)
            elif entry.is_dir() and not entry.name.startswith("."):
                result = count_files(os.path.join(path, entry.name), visitor, result)
    return result


def file_counter(result, path, name):
    return result + 1


def loc(file):
    try:
        with open(file, "r", 1, "utf-8") as f:
            return len(f.readlines())
    except:
        return 0


def loc_counter(result, path, name):
    return result + loc(os.path.join(path, name))


def composite_counter(result, path, name):
    result["file"] = file_counter(result["file"], path, name)
    result["loc"] = loc_counter(result["loc"], path, name)
    return result


def oldest_commit_between_dates(git: instarepo.git.GitWorkingDir, since, until):
    lines = git.rev_list(since, until, reverse=True)
    return lines[0] if len(lines) > 0 else ""



import logging
import os.path

import instarepo.git
import instarepo.github
import instarepo.repo_source


class CloneCommand:
    def __init__(self, args):
        self.repo_source = (
            instarepo.repo_source.RepoSourceBuilder().with_args(args).build()
        )
        self.verbose: bool = args.verbose
        self.projects_dir: str = args.projects_dir

    def run(self):
        if not os.path.isdir(self.projects_dir):
            raise ValueError(f"Projects dir {self.projects_dir} does not exist")
        repos = self.repo_source.get()
        for repo in repos:
            project_dir = os.path.join(self.projects_dir, repo.name)
            if os.path.isdir(project_dir):
                logging.info("Skipping %s because it already exists", repo.name)
            else:
                logging.info("Cloning %s", repo.name)
                instarepo.git.clone(repo.ssh_url, project_dir, quiet=not self.verbose)



"""Main entrypoint of the program"""
import argparse
import logging

import instarepo.commands.analyze
import instarepo.commands.clone
import instarepo.commands.fix
import instarepo.commands.list
import instarepo.commands.login
import instarepo.commands.logout

VERSION = "0.8.1"


def main():
    """Main entrypoint of the program"""
    args = parse_args()
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        level=logging.DEBUG if args.verbose else logging.INFO,
    )
    cmd = create_command(args)
    cmd.run()


def create_command(args):
    """Creates the command handler for the given parsed CLI arguments"""
    if args.subparser_name == "analyze":
        cmd = instarepo.commands.analyze.AnalyzeCommand(args)
    elif args.subparser_name == "fix":
        cmd = instarepo.commands.fix.FixCommand(args)
    elif args.subparser_name == "list":
        cmd = instarepo.commands.list.ListCommand(args)
    elif args.subparser_name == "clone":
        cmd = instarepo.commands.clone.CloneCommand(args)
    elif args.subparser_name == "login":
        cmd = instarepo.commands.login.LoginCommand(args)
    elif args.subparser_name == "logout":
        cmd = instarepo.commands.logout.LogoutCommand(args)
    else:
        raise ValueError(f"Sub-parser {args.subparser_name} is not implemented")
    return cmd


def parse_args(args=None):
    """Parses the CLI arguments"""
    parser = argparse.ArgumentParser(
        description="Apply changes on multiple repositories", prog="instarepo"
    )
    parser.add_argument(
        "--verbose", action="store_true", default=False, help="Verbose output"
    )
    parser.add_argument("--version", action="version", version="%(prog)s " + VERSION)

    subparsers = parser.add_subparsers(
        dest="subparser_name", help="Sub-commands help", required=True
    )

    list_parser = subparsers.add_parser(
        "list",
        help="Lists the available repositories",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""Example:
    instarepo list -u USER -t TOKEN
    """,
    )
    _configure_list_parser(list_parser)

    fix_parser = subparsers.add_parser(
        "fix",
        description="Runs automatic fixes on the repositories",
        help="Runs automatic fixes on the repositories",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""Example:
    instarepo fix -u USER -t TOKEN

Fixers:

"""
        + instarepo.commands.fix.epilog(),
    )
    _configure_fix_parser(fix_parser)

    analyze_parser = subparsers.add_parser(
        "analyze",
        help="Analyzes the available repositories, counting historical LOC",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""Example:
    instarepo analyze -u USER -t TOKEN --since 2021-11-06
    """,
    )
    _configure_analyze_parser(analyze_parser)

    clone_parser = subparsers.add_parser(
        "clone",
        help="Clones all the available repositories",
    )
    _configure_clone_parser(clone_parser)

    login_parser = subparsers.add_parser(
        "login",
        help="Provide GitHub credentials for subsequent commands",
    )
    _configure_login_parser(login_parser)

    subparsers.add_parser(
        "logout",
        help="Delete previously stored GitHub credentials",
    )

    return parser.parse_args(args)


def _configure_list_parser(parser: argparse.ArgumentParser):
    _add_auth_options(parser)
    _add_sort_options(parser)
    _add_filter_options(parser)
    _add_archived_option(parser)


def _configure_fix_parser(parser: argparse.ArgumentParser):
    _add_auth_options(parser)
    _add_sort_options(parser)
    _add_filter_options(parser)
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not actually push and create MR",
    )
    fixer_group = parser.add_mutually_exclusive_group()
    fixer_group.add_argument(
        "--only-fixers",
        action="store",
        nargs="+",
        help="Only run fixers that have the given prefixes",
    )
    fixer_group.add_argument(
        "--except-fixers",
        action="store",
        nargs="+",
        help="Do not run fixers that have the given prefixes",
    )
    parser.add_argument(
        "--local-dir",
        help="Apply fixes for a project at a local working directory. Skips all GitHub related calls and git push.",
    )


def _configure_analyze_parser(parser: argparse.ArgumentParser):
    _add_auth_options(parser)
    _add_sort_options(parser)
    _add_filter_options(parser)
    _add_archived_option(parser)
    parser.add_argument(
        "--since",
        required=True,
        action="store",
        help="The start date of the analysis (YYYY-mm-dd)",
    )
    parser.add_argument(
        "--metric",
        choices=["commits", "files"],
        default="commits",
        help="The metric to report on",
    )


def _configure_clone_parser(parser: argparse.ArgumentParser):
    _add_auth_options(parser)
    _add_archived_option(parser)
    _add_filter_options(parser)
    parser.add_argument(
        "--projects-dir",
        required=True,
        help="The directory where projects are going to be cloned into",
    )


def _configure_login_parser(parser: argparse.ArgumentParser):
    _add_auth_options(parser, required=True)


def _add_archived_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--archived",
        action="store",
        default="deny",
        choices=["allow", "deny", "only"],
        help="Filter archived repositories",
    )


def _add_auth_options(parser: argparse.ArgumentParser, required=False):
    auth_group = parser.add_argument_group("Authentication")
    auth_group.add_argument(
        "-u", "--username", required=required, help="The GitHub username"
    )
    auth_group.add_argument("-t", "--token", required=required, help="The GitHub token")


def _add_sort_options(parser: argparse.ArgumentParser):
    sort_group = parser.add_argument_group("Sorting")
    sort_group.add_argument(
        "--sort",
        action="store",
        default="full_name",
        choices=["full_name", "created", "updated", "pushed"],
    )
    sort_group.add_argument(
        "--direction", action="store", default="asc", choices=["asc", "desc"]
    )


def _add_filter_options(parser: argparse.ArgumentParser):
    language_group = parser.add_mutually_exclusive_group()
    language_group.add_argument(
        "--only-language",
        help="Only process repositories of the given programming language",
    )
    language_group.add_argument(
        "--except-language",
        help="Do not process repositories of the given programming language",
    )

    prefix_group = parser.add_mutually_exclusive_group()
    prefix_group.add_argument(
        "--only-name-prefix",
        help="Only process repositories whose name starts with the given prefix",
    )
    prefix_group.add_argument(
        "--except-name-prefix",
        help="Do not process repositories whose name starts with the given prefix",
    )

    filter_group = parser.add_argument_group("Filtering")

    filter_group.add_argument(
        "--forks",
        default="deny",
        choices=["allow", "deny", "only"],
        help="Filter forks",
    )
    filter_group.add_argument(
        "--pushed-after",
        help="Only process repositories that had changes pushed after the given time interval e.g. 4h",
    )
    filter_group.add_argument(
        "--pushed-before",
        help="Only process repositories that had changes pushed before the given time interval e.g. 4h",
    )


if __name__ == "__main__":
    main()



import configparser
import os
import os.path
import pathlib
import requests.auth


class Credentials:
    def __init__(self, username, token):
        self.username = username
        self.token = token

    def is_valid(self):
        return self.username and self.token

    def with_fallback(self, fallback):
        return Credentials(
            self.username or fallback.username, self.token or fallback.token
        )


if os.name == "nt":
    import winreg

    class WindowsCredentials:
        def __init__(self):
            pass

        def load(self):
            with winreg.CreateKey(winreg.HKEY_CURRENT_USER, "SOFTWARE") as k1:
                with winreg.CreateKey(k1, "ngeor") as k2:
                    with winreg.CreateKey(k2, "instarepo") as k3:
                        username, _ = winreg.QueryValueEx(k3, "username")
                        token, _ = winreg.QueryValueEx(k3, "token")
                        return Credentials(username, token)

        def store(self, credentials: Credentials):
            username = credentials.username
            token = credentials.token
            with winreg.CreateKey(winreg.HKEY_CURRENT_USER, "SOFTWARE") as k1:
                with winreg.CreateKey(k1, "ngeor") as k2:
                    with winreg.CreateKey(k2, "instarepo") as k3:
                        winreg.SetValueEx(k3, "username", None, winreg.REG_SZ, username)
                        winreg.SetValueEx(k3, "token", None, winreg.REG_SZ, token)

        def clear(self):
            with winreg.CreateKey(winreg.HKEY_CURRENT_USER, "SOFTWARE") as k1:
                with winreg.CreateKey(k1, "ngeor") as k2:
                    with winreg.CreateKey(k2, "instarepo") as k3:
                        winreg.DeleteValue(k3, "username")
                        winreg.DeleteValue(k3, "token")


class DotFileCredentials:
    def __init__(self):
        self.file = os.path.join(pathlib.Path.home(), ".instarepo.ini")

    def load(self):
        username = ""
        token = ""
        if os.path.isfile(self.file):
            config = configparser.ConfigParser()
            config.read(self.file)
            if "DEFAULT" in config:
                default_section = config["DEFAULT"]
                username = default_section.get("username", "")
                token = default_section.get("token", "")
        return Credentials(username, token)

    def store(self, credentials: Credentials):
        config = configparser.ConfigParser()
        config["DEFAULT"] = {
            "username": credentials.username,
            "token": credentials.token,
        }
        with open(self.file, "w") as file:
            config.write(file)

    def clear(self):
        if os.path.isfile(self.file):
            os.remove(self.file)


class PlatformCredentials:
    def __init__(self):
        if os.name == "nt":
            self.delegate = WindowsCredentials()
        else:
            self.delegate = DotFileCredentials()

    def load(self):
        return self.delegate.load()

    def store(self, credentials: Credentials):
        self.delegate.store(credentials)

    def clear(self):
        self.delegate.clear()


def load_credentials(args):
    args_credentials = Credentials(args.username, args.token)
    if args_credentials.is_valid():
        return args_credentials

    stored_credentials = PlatformCredentials().load()
    mixed_credentials = args_credentials.with_fallback(stored_credentials)
    if not mixed_credentials.is_valid():
        raise ValueError(
            "Credentials not provided. Use the CLI parameters or login first."
        )
    return mixed_credentials


def build_requests_auth(args):
    credentials = load_credentials(args)
    return requests.auth.HTTPBasicAuth(credentials.username, credentials.token)



import datetime
import os.path
import subprocess


class GitWorkingDir:
    def __init__(self, directory: str, quiet: bool = False):
        self.dir = directory
        self.quiet = quiet

    def join(self, *args) -> str:
        root = self.dir
        for arg in args:
            parts = arg.replace("\\", "/").split("/")
            for part in parts:
                root = os.path.join(root, part)
        return root

    def isfile(self, *args) -> bool:
        return os.path.isfile(self.join(*args))

    def isdir(self, *args) -> bool:
        return os.path.isdir(self.join(*args))

    def create_branch(self, name: str) -> None:
        args = ["git", "checkout"]
        if self.quiet:
            args.append("-q")
        args.extend(["-b", name])
        subprocess.run(
            args,
            check=True,
            cwd=self.dir,
        )

    def checkout(self, name: str) -> None:
        args = ["git", "checkout"]
        if self.quiet:
            args.append("-q")
        args.append(name)
        subprocess.run(
            args,
            check=True,
            cwd=self.dir,
        )

    def add(self, file: str) -> None:
        subprocess.run(["git", "add", file], check=True, cwd=self.dir)

    def rm(self, file: str) -> None:
        subprocess.run(["git", "rm", file], check=True, cwd=self.dir)

    def commit(self, message: str) -> None:
        args = ["git", "commit"]
        if self.quiet:
            args.append("-q")
        args.extend(["-m", message])
        subprocess.run(args, check=True, cwd=self.dir)

    def push(self, force: bool = False, remote_name: str = "origin") -> None:
        args = ["git", "push"]
        if self.quiet:
            args.append("-q")
        if force:
            args.append("--force-with-lease")
        args.extend(["-u", remote_name, "HEAD"])
        subprocess.run(
            args,
            check=True,
            cwd=self.dir,
        )

    def delete_local_branch(self, branch_name: str) -> None:
        args = ["git", "branch"]
        if self.quiet:
            args.append("-q")
        args.extend(["-D", branch_name])
        subprocess.run(
            args,
            check=True,
            cwd=self.dir,
        )

    def delete_remote_branch(
        self, branch_name: str, remote_name: str = "origin"
    ) -> None:
        args = ["git", "push"]
        if self.quiet:
            args.append("-q")
        args.extend(["--delete", remote_name, branch_name])
        subprocess.run(
            args,
            check=True,
            cwd=self.dir,
        )

    def rev_parse(self, branch_name: str) -> str:
        """
        Gets the SHA of the given branch.
        """
        result = subprocess.run(
            ["git", "rev-parse", "-q", "--verify", branch_name],
            check=True,
            cwd=self.dir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        return result.stdout.strip()

    def current_branch_name(self) -> str:
        result = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            check=True,
            cwd=self.dir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        return result.stdout.strip()

    def user_name(self) -> str:
        """
        Gets the `user.name` configured property.
        """
        result = subprocess.run(
            ["git", "config", "user.name"],
            check=True,
            cwd=self.dir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        return result.stdout.strip()

    def rev_list(self, since, until, reverse=False):
        args = ["git", "rev-list", f"--since={since}", f"--until={until}"]
        if reverse:
            args.append("--reverse")
        args.append("HEAD")
        result = subprocess.run(
            args, check=True, cwd=self.dir, encoding="utf-8", stdout=subprocess.PIPE
        )
        return result.stdout.splitlines()

    def commit_date(self, commit_id):
        result = subprocess.run(
            ["git", "show", "-s", "--format=%ci", commit_id],
            check=True,
            cwd=self.dir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )

        # the result looks like 2017-02-20 21:28:35 +0100
        return datetime.datetime.strptime(
            result.stdout.split(" ")[0], "%Y-%m-%d"
        ).date()

    def is_remote_branch_present(self, branch: str, remote="origin"):
        remote_branch_sha = ""
        try:
            remote_branch_sha = self.rev_parse(f"remotes/origin/{branch}")
        except:  # pylint: disable=bare-except
            pass
        return len(remote_branch_sha) > 0

    def get_behind_ahead(self, base: str, head: str):
        # git rev-list --left-right --count origin/trunk...trunk
        # 12      0
        result = subprocess.run(
            ["git", "rev-list", "--left-right", "--count", f"{base}...{head}"],
            check=True,
            cwd=self.dir,
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )
        parts = result.stdout.split("\t")
        return int(parts[0]), int(parts[1])


def clone(ssh_url: str, clone_dir: str, quiet: bool = False) -> GitWorkingDir:
    args = ["git", "clone"]
    if quiet:
        args.append("-q")
    args.extend([ssh_url, clone_dir])
    subprocess.run(args, check=True)
    return GitWorkingDir(clone_dir, quiet)



import datetime
import os.path
import re

from typing import Optional

import instarepo.git
import instarepo.github
from instarepo.fixers.base import MissingFileFix


def _repl(m: re.Match, year: int) -> str:
    is_year_range = m.group(4)
    if is_year_range:
        is_same_year = str(year) == m.group(4)
        if is_same_year:
            return m.group(0)
        else:
            return m.group(1) + m.group(2) + "-" + str(year)
    else:
        is_same_year = str(year) == m.group(2)
        if is_same_year:
            return m.group(0)
        else:
            return m.group(0) + "-" + str(year)


def update_copyright_year(contents: str, year: int) -> str:
    x = re.compile(r"^(Copyright \(c\) )([0-9]{4})(-([0-9]{4}))?", re.M)
    return x.sub(lambda m: _repl(m, year), contents)


class CopyrightYearFix:
    """
    Ensures the year in the license file copyright is up to date.

    Does not run for forks, private repos, and local git repos.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        repo: Optional[instarepo.github.Repo],
        **kwargs
    ):
        self.git = git
        self.repo = repo

    def run(self):
        if not self.repo or self.repo.private or self.repo.fork:
            return []
        filename = os.path.join(self.git.dir, "LICENSE")
        if not os.path.isfile(filename):
            return []
        with open(filename, "r", encoding="utf-8") as f:
            old_contents = f.read()
        new_contents = update_copyright_year(old_contents, datetime.date.today().year)
        if old_contents == new_contents:
            return []
        with open(filename, "w", encoding="utf8") as f:
            f.write(new_contents)
        self.git.add("LICENSE")
        msg = "chore: Updated copyright year in LICENSE"
        self.git.commit(msg)
        return [msg]


MIT_LICENSE = """MIT License

Copyright (c) [year] [fullname]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class MustHaveLicenseFix(MissingFileFix):
    """
    Ensures that a license file exists.

    Does not run for forks, private repos, and local git repos.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        repo: Optional[instarepo.github.Repo],
        **kwargs
    ):
        super().__init__(git, "LICENSE")
        self.repo = repo

    def should_process_repo(self):
        return self.repo and not self.repo.private and not self.repo.fork

    def get_contents(self):
        contents = MIT_LICENSE.replace(
            "[year]", str(datetime.date.today().year)
        ).replace("[fullname]", self.git.user_name())
        return contents



import logging
import os.path
import re
from typing import List, Optional

import instarepo.git
import instarepo.github


RE_DESCRIPTION_LINE = re.compile(r"^[>A-Za-z]")


class RepoDescriptionFix:
    """
    Updates GitHub's repo description based on the README file.

    Note: this fixer does not create an MR, it calls the
    GitHub REST API directly (https://docs.github.com/en/rest/reference/repos#update-a-repository).

    Does not run for local git repositories.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        github: Optional[instarepo.github.GitHub],
        repo: Optional[instarepo.github.Repo],
        **kwargs
    ):
        self.github = github
        self.git = git
        self.repo = repo

    def run(self):
        readme_description = self.get_readme_description()
        if not readme_description or not self.repo or not self.github:
            return []
        if readme_description != self.repo.description:
            logging.info(
                "Repo description %s does not match readme description %s",
                self.repo.description,
                readme_description,
            )
            self.github.update_description(self.repo.full_name, readme_description)
        # this fixer does not create an MR
        return []

    def get_readme_description(self):
        filename = os.path.join(self.git.dir, "README.md")
        if not os.path.isfile(filename):
            return None
        with open(filename) as f:
            # read lines
            lines = f.readlines()
            # trim
            lines = [line.strip() for line in lines]
            # keep the ones that start with a letter
            lines = [line for line in lines if RE_DESCRIPTION_LINE.match(line)]
            return get_description_from_lines(lines)


def get_description_from_lines(lines: List[str]) -> str:
    if not lines:
        return ""
    line = lines[0]
    if line.startswith(">"):
        line = line[1:].strip()
    return line



"""XML utilities"""
import xml.etree.ElementTree as ET
from typing import Optional


def create_parser():
    """Creates an XML parser that is aware of comments"""
    return ET.XMLParser(target=ET.TreeBuilder(insert_comments=True))


def parse(filename: str):
    """Parses the given XML file"""
    return ET.parse(filename, parser=create_parser())


def find_at_tree(tree: ET.ElementTree, *args):
    if tree is None:
        return None
    return find(tree.getroot(), *args)


def find(node: Optional[ET.Element], *args) -> Optional[ET.Element]:
    p = node
    for arg in args:
        if p is None:
            return None
        p = p.find(arg)
    return p



import os.path
import requests
import instarepo.git
import instarepo.github
from instarepo.fixers.base import MissingFileFix
from .finders import is_lazarus_project, is_maven_project, is_vb6_project


class MustHaveReadmeFix(MissingFileFix):
    """Ensures that the repo has a readme file"""

    def __init__(
        self, git: instarepo.git.GitWorkingDir, repo: instarepo.github.Repo, **kwargs
    ):
        super().__init__(git, "README.md")
        self.repo = repo

    def get_contents(self):
        contents = f"# {self.repo.name}\n"
        if self.repo.description:
            contents = contents + "\n" + self.repo.description + "\n"
        return contents


EDITOR_CONFIG = """# Editor configuration, see https://editorconfig.org
root = true

[*]
charset = utf-8
indent_style = space
indent_size = 4
insert_final_newline = true
trim_trailing_whitespace = true
max_line_length = 120

[*.sh]
end_of_line = lf
"""


class MustHaveEditorConfigFix(MissingFileFix):
    """Ensures an editorconfig file exists"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        super().__init__(git, ".editorconfig")

    def get_contents(self):
        return EDITOR_CONFIG


class MustHaveGitHubFundingFix(MissingFileFix):
    """
    Ensures a GitHub funding file exists (.github/FUNDING.yml).
    The rule will use the FUNDING.yml file from `user-templates/.github/FUNDING.yml`,
    if one exists.
    """

    def __init__(
        self, git: instarepo.git.GitWorkingDir, repo: instarepo.github.Repo, **kwargs
    ):
        super().__init__(git, ".github/FUNDING.yml")
        self.repo = repo
        template = os.path.join(
            os.path.basename(__file__),
            "..",
            "..",
            "user-templates",
            ".github",
            "FUNDING.yml",
        )
        self.contents = ""
        if os.path.isfile(template):
            with open(template, "r", encoding="utf-8") as file:
                self.contents = file.read()

    def should_process_repo(self):
        return not self.repo.private and not self.repo.fork and self.contents

    def get_contents(self):
        return self.contents


class MustHaveGitIgnoreFix(MissingFileFix):
    """Ensures a .gitignore file exists"""

    LAZARUS_GITIGNORE = """*.o
*.ppu
*.obj
*.exe
*.dll
*.compiled
*.bak
*.lps
backup/
"""
    VB6_GITIGNORE = """*.exe
*.dll
*.ocx
*.vbw
"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        super().__init__(git, ".gitignore")

    def get_contents(self):
        if is_maven_project(self.git.dir):
            # https://github.com/github/gitignore/blob/master/Maven.gitignore
            response = requests.get(
                "https://raw.githubusercontent.com/github/gitignore/master/Maven.gitignore"
            )
            response.raise_for_status()
            return response.text
        if is_lazarus_project(self.git.dir):
            return self.LAZARUS_GITIGNORE
        if is_vb6_project(self.git.dir):
            return self.VB6_GITIGNORE
        return None



"""
Functions that determine the project's language or build system.
"""

import os
import os.path


def is_lazarus_project(directory: str) -> bool:
    """
    Checks if the given directory is a Lazarus project.
    """
    return has_file_of_extension(directory, ".lpr")


def is_maven_project(directory: str) -> bool:
    """
    Checks if the given directory is a Maven project.
    """
    return os.path.isfile(os.path.join(directory, "pom.xml"))


def is_vb6_project(directory: str) -> bool:
    """
    Checks if the given directory is a VB6 project.
    """
    return has_file_of_extension(directory, ".vbp", ".vbg")


def has_file_of_extension(directory: str, *args) -> bool:
    """
    Checks if the given directory has a file of the
    given file extensions.

    Extensions are passed in the args parameter and
    must be prefixed by a dot.
    """
    with os.scandir(directory) as iterator:
        for entry in iterator:
            if is_file_of_extension(entry, *args):
                return True
    return False


def is_file_of_extension(entry, *args):
    """
    Checks if the given directory entry is a file
    and ends in one of the given extensions.

    Extensions are passed in the args parameter and
    must be prefixed by a dot.
    """
    if not entry.is_file():
        return False
    if not args:
        return True
    for arg in args:
        if entry.name.endswith(arg):
            return True
    return False



from datetime import timedelta
from .github import Repo
from .repo_source import (
    StringFilter,
    FilterMode,
    filter_by_name_prefix,
    parse_timedelta,
)


def test_filter_by_repo_no_filter():
    repos = [dummy_repo("foo"), dummy_repo("bar")]
    result = list(filter_by_name_prefix(repos, None))
    assert result == repos


def test_filter_by_repo_allow_filter():
    repos = [dummy_repo("foo"), dummy_repo("bar")]
    string_filter = StringFilter("whatever", FilterMode.ALLOW)
    result = list(filter_by_name_prefix(repos, string_filter))
    assert result == repos


def test_filter_by_repo_only_filter():
    repos = [dummy_repo("foo"), dummy_repo("bar")]
    string_filter = StringFilter("f", FilterMode.ONLY)
    result = list(filter_by_name_prefix(repos, string_filter))
    assert result == repos[0:1]


def test_filter_by_repo_deny_filter():
    repos = [dummy_repo("foo"), dummy_repo("bar")]
    string_filter = StringFilter("f", FilterMode.DENY)
    result = list(filter_by_name_prefix(repos, string_filter))
    assert result == repos[1:]


def dummy_repo(name: str) -> Repo:
    return Repo(
        {
            "name": name,
            "archived": False,
            "clone_url": "",
            "html_url": "",
            "ssh_url": "",
            "default_branch": "",
            "full_name": "",
            "description": "",
            "private": False,
            "fork": False,
            "created_at": "2021-11-04T21:32:00Z",
            "pushed_at": "2021-11-04T21:32:00Z",
            "updated_at": "2021-11-04T21:32:00Z",
            "language": "",
        }
    )


def test_parse_timedelta():
    assert parse_timedelta("4h") == timedelta(hours=4)
    assert parse_timedelta("15m") == timedelta(minutes=15)
    assert parse_timedelta("3d") == timedelta(days=3)
    assert parse_timedelta(None) == None



"""Fixers for .NET projects"""
import logging
import os
import os.path
import xml.etree.ElementTree as ET
from typing import Iterable, List, Optional

import instarepo.git
import instarepo.github
import instarepo.xml_utils
from .base import ensure_directories
from .finders import is_file_of_extension


class DotNetFrameworkVersionFix:
    """Sets the .NET Framework version to 4.7.2 in csproj and web.config files"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        self.git = git
        self.result = []

    def run(self):
        # ensure abspath so that dirpath in the loop is also absolute
        self.result = []
        with os.scandir(self.git.dir) as iterator:
            for entry in iterator:
                if is_file_of_extension(entry, ".sln"):
                    self.process_sln(entry.path)
        return self.result

    def process_sln(self, sln_path: str):
        for relative_csproj in get_projects_from_sln_file(sln_path):
            parts = relative_csproj.replace("\\", "/").split("/")
            abs_csproj = os.path.join(self.git.dir, *parts)
            self.process_csproj(abs_csproj)
            directory_parts = parts[0:-1]
            csproj_dir = os.path.join(self.git.dir, *directory_parts)
            self.process_web_configs(csproj_dir)

    def process_web_configs(self, csproj_dir: str):
        for web_config in get_web_configs_from_dir(csproj_dir):
            self.process_web_config(web_config)

    def process_csproj(self, filename: str):
        logging.debug("Processing csproj %s", filename)
        ET.register_namespace("", "http://schemas.microsoft.com/developer/msbuild/2003")
        try:
            tree = instarepo.xml_utils.parse(filename)
            target_framework_version = instarepo.xml_utils.find_at_tree(
                tree,
                "{http://schemas.microsoft.com/developer/msbuild/2003}PropertyGroup",
                "{http://schemas.microsoft.com/developer/msbuild/2003}TargetFrameworkVersion",
            )
            if target_framework_version is None:
                return
            desired_framework_version = "v4.7.2"
            if target_framework_version.text == desired_framework_version:
                return
            target_framework_version.text = desired_framework_version
            tree.write(
                filename,
                xml_declaration=True,
                encoding="utf-8",
            )
        finally:
            ET.register_namespace(
                "msbuild", "http://schemas.microsoft.com/developer/msbuild/2003"
            )
        relpath = os.path.relpath(filename, self.git.dir)
        self.git.add(relpath)
        msg = f"chore: Upgraded {relpath} to .NET {desired_framework_version}"
        self.git.commit(msg)
        self.result.append(msg)

    def process_web_config(self, filename: str):
        logging.debug("Processing web.config %s", filename)
        tree = instarepo.xml_utils.parse(filename)
        compilation = instarepo.xml_utils.find_at_tree(
            tree, "system.web", "compilation"
        )
        if compilation is None:
            return
        desired_framework_version = "4.7.2"
        if compilation.attrib.get("targetFramework", "") == desired_framework_version:
            return
        compilation.attrib["targetFramework"] = desired_framework_version
        tree.write(
            filename,
            xml_declaration=True,
            encoding="utf-8",
        )
        relpath = os.path.relpath(filename, self.git.dir)
        self.git.add(relpath)
        msg = f"chore: Upgraded {relpath} to .NET {desired_framework_version}"
        self.git.commit(msg)
        self.result.append(msg)


class MustHaveGitHubActionFix:
    """
    Creates a GitHub Action workflow for CSharp projects, deletes appveyor.yml if present.
    Does not work for locally checked out repositories.
    """

    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        repo: Optional[instarepo.github.Repo],
        **kwargs,
    ):
        self.git = git
        self.repo = repo

    def run(self):
        if not self._should_process_repo():
            return []

        if not self.repo:
            return []

        expected_contents = get_workflow_contents(self.repo)
        dir_name = ".github/workflows"
        ensure_directories(self.git, dir_name)
        file_name = dir_name + "/build.yml"
        absolute_file_name = self.git.join(file_name)
        if os.path.isfile(absolute_file_name):
            with open(absolute_file_name, "r", encoding="utf-8") as file:
                old_contents = file.read()
        else:
            old_contents = ""
        if expected_contents != old_contents:
            with open(absolute_file_name, "w", encoding="utf-8") as file:
                file.write(expected_contents)
            self.git.add(file_name)
            if old_contents:
                msg = "chore: Updated GitHub Actions workflow for .NET project"
            else:
                msg = "chore: Added GitHub Actions workflow for .NET project"
            self._rm_appveyor()
            self.git.commit(msg)
            return [msg]
        if self._rm_appveyor():
            msg = "chore: Removed appveyor.yml from .NET project"
            self.git.commit(msg)
            return [msg]
        return []

    def _should_process_repo(self) -> bool:
        """
        Checks if the repo should be processed.
        The repo should be processed if it contains exactly one sln file
        at the root directory which references at least one csproj file.
        """
        sln_path = ""
        with os.scandir(self.git.dir) as iterator:
            for entry in iterator:
                if is_file_of_extension(entry, ".sln"):
                    if sln_path:
                        # multiple sln files not supported currently
                        return False
                    else:
                        sln_path = entry.path
        if not sln_path:
            return False
        return len(get_projects_from_sln_file(sln_path)) > 0

    def _rm_appveyor(self):
        if self.git.isfile("appveyor.yml"):
            self.git.rm("appveyor.yml")
            return True
        return False


def get_workflow_contents(repo: instarepo.github.Repo):
    return """name: CI

on:
  push:
    branches: [ trunk ]
  pull_request:
    branches: [ trunk ]

jobs:
  build:
    runs-on: ubuntu-latest
    steps:
    - uses: actions/checkout@v2
    - name: Set up .NET
      uses: actions/setup-dotnet@v1
      with:
        dotnet-version: '3.1.x'
    - run: dotnet build
    - run: dotnet test
""".replace(
        "trunk", repo.default_branch
    )


def get_projects_from_sln_file(path: str) -> List[str]:
    """
    Gets the projects defined in a sln file.

    :param path: The path of a Visual Studio sln file.
    """
    with open(path, "r", encoding="utf-8") as file:
        return list(get_projects_from_sln_file_contents(file.read()))


def get_projects_from_sln_file_contents(contents: str) -> Iterable[str]:
    """
    Gets the projects defined in a sln file.

    :param contents: The contents of a Visual Studio sln file.
    """
    for line in contents.splitlines():
        if line.startswith('Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}")'):
            parts = line.split(",")
            csproj = parts[1].strip()
            if csproj.startswith('"'):
                csproj = csproj[1:].strip()
            if csproj.endswith('"'):
                csproj = csproj[:-1].strip()
            yield csproj


def get_web_configs_from_dir(csproj_dir: str) -> Iterable[str]:
    """
    Gets the files named web.config in a directory.
    In a case-sensitive file system, if the directory contains both
    web.config and Web.Config, both will be returned.
    The iterator yields absolute paths.
    """
    with os.scandir(csproj_dir) as iterator:
        for entry in iterator:
            if entry.is_file() and entry.name.lower() == "web.config":
                yield entry.path



"""
Unit tests for ci.py
"""

from .ci import remove_travis_badge


def test_badge():
    """Tests removing the Travis badge"""
    x = "[![Build Status](https://travis-ci.org/ngeor/games.svg?branch=master)](https://travis-ci.org/ngeor/games)"
    assert remove_travis_badge(x) == ""

    x = "hi"
    assert remove_travis_badge(x) == x



"""
Fixers regarding CI.
"""

import os.path
import re
import instarepo.fixers.base
import instarepo.git


class NoTravisFix:
    """Removes the .travis.yml file"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        self.git = git

    def run(self):
        filename = ".travis.yml"
        full_name = os.path.join(self.git.dir, filename)
        if not os.path.isfile(full_name):
            return []
        self.git.rm(filename)
        msg = f"chore: Removed {filename}"
        self.git.commit(msg)
        return [msg]


class NoTravisBadgeFix(instarepo.fixers.base.SingleFileFix):
    """Removes the Travis badge from README files"""

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        super().__init__(git, "README.md", "chore: Removed Travis badge from README")

    def convert(self, contents: str) -> str:
        return remove_travis_badge(contents)


RE_BADGE = re.compile(
    r"\[!\[Build Status\]\(https://travis-ci[^)]+\)\]\(https://travis-ci[^)]+\)"
)


def remove_travis_badge(contents: str) -> str:
    return RE_BADGE.sub("", contents)



import datetime
import logging
import requests


class Repo:
    def __init__(self, repo_json):
        self.name: str = repo_json["name"]
        self.archived: bool = repo_json["archived"]
        self.clone_url: str = repo_json["clone_url"]
        self.html_url: str = repo_json["html_url"]
        self.ssh_url: str = repo_json["ssh_url"]
        self.default_branch: str = repo_json["default_branch"]
        self.full_name: str = repo_json["full_name"]
        self.description: str = repo_json["description"]
        self.private: bool = repo_json["private"]
        self.fork: bool = repo_json["fork"]
        self.created_at = datetime.datetime.strptime(
            repo_json["created_at"], "%Y-%m-%dT%H:%M:%S%z"
        )
        self.pushed_at = datetime.datetime.strptime(
            repo_json["pushed_at"], "%Y-%m-%dT%H:%M:%S%z"
        )
        self.updated_at = datetime.datetime.strptime(
            repo_json["updated_at"], "%Y-%m-%dT%H:%M:%S%z"
        )
        self.language: str = repo_json["language"]


class GitHub:
    """
    A read-only GitHub client.
    """

    def __init__(self, auth):
        self.auth = auth

    def get_all_repos(self, sort: str, direction: str):
        page = 1
        per_page = 30
        has_more = True
        while has_more:
            count = 0
            for repo in self.get_all_repos_of_page(sort, direction, page, per_page):
                count = count + 1
                yield repo
            page = page + 1
            has_more = count >= per_page

    def get_all_repos_of_page(
        self, sort: str, direction: str, page: int, per_page: int
    ):
        # https://docs.github.com/en/rest/reference/repos#list-repositories-for-the-authenticated-user
        response = requests.get(
            "https://api.github.com/user/repos",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            params={
                "sort": sort,
                "direction": direction,
                "page": page,  # first page's index is 1
                "per_page": per_page,  # Default: 30
                "visibility": "all",  # Can be one of all, public, or private. Default: all
                "affiliation": "owner",  # Comma-separated list of values. Default: owner,collaborator,organization_member
            },
        )
        response.raise_for_status()
        for repo in response.json():
            logging.debug("yield repo: %s", repo)
            yield Repo(repo)

    def create_merge_request(
        self, full_name: str, head: str, base: str, title: str, body: str
    ) -> str:
        logging.debug("Would have created MR")
        return ""

    def update_description(self, full_name: str, description: str):
        logging.debug("Would have set description of %s to %s", full_name, description)

    def list_merge_requests(self, full_name: str, head: str, base: str):
        # https://docs.github.com/en/rest/reference/pulls#list-pull-requests
        response = requests.get(
            f"https://api.github.com/repos/{full_name}/pulls",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            params={"head": head, "base": base},
        )
        result = response.json()
        response.raise_for_status()
        return result

    def get_merge_request(self, full_name: str, pull_number: int):
        href = f"https://api.github.com/repos/{full_name}/pulls/{pull_number}"
        return self.get_json(href)

    def close_merge_request(self, full_name: str, pull_number: int):
        logging.info("Would have closed MR %s %d", full_name, pull_number)

    def merge_merge_request(self, full_name: str, pull_number: int):
        logging.info("Would have merged MR %s %d", full_name, pull_number)

    def get_json(self, href: str):
        response = requests.get(
            href, auth=self.auth, headers={"Accept": "application/vnd.github.v3+json"}
        )
        result = response.json()
        response.raise_for_status()
        return result

    def create_issue_comment(self, full_name: str, issue_number: int, body: str):
        logging.info("Would have created issue comment %s %d", full_name, issue_number)

    def list_check_runs(self, full_name: str, sha: str):
        # https://docs.github.com/en/rest/reference/checks#list-check-runs-for-a-git-reference
        return self.get_json(
            f"https://api.github.com/repos/{full_name}/commits/{sha}/check-runs"
        )


class ReadWriteGitHub(GitHub):
    """
    A read-write GitHub client.
    """

    def __init__(self, auth):
        super().__init__(auth)

    def create_merge_request(
        self, full_name: str, head: str, base: str, title: str, body: str
    ) -> str:
        # https://docs.github.com/en/rest/reference/pulls#create-a-pull-request
        response = requests.post(
            f"https://api.github.com/repos/{full_name}/pulls",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            json={
                "head": head,
                "base": base,
                "title": title,
                "body": body,
            },
        )
        result = response.json()
        response.raise_for_status()
        return result["html_url"]

    def update_description(self, full_name: str, description: str):
        # https://docs.github.com/en/rest/reference/repos#update-a-repository
        response = requests.patch(
            f"https://api.github.com/repos/{full_name}",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            json={"description": description},
        )
        response.raise_for_status()

    def close_merge_request(self, full_name: str, pull_number: int):
        # https://docs.github.com/en/rest/reference/pulls#update-a-pull-request
        logging.info("Closing PR %s %d", full_name, pull_number)
        response = requests.patch(
            f"https://api.github.com/repos/{full_name}/pulls/{pull_number}",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            json={"state": "closed"},
        )
        response.raise_for_status()

    def merge_merge_request(self, full_name: str, pull_number: int):
        # https://docs.github.com/en/rest/reference/pulls#merge-a-pull-request
        logging.info("Merging PR %s %d", full_name, pull_number)
        response = requests.put(
            f"https://api.github.com/repos/{full_name}/pulls/{pull_number}/merge",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
        )
        response.raise_for_status()

    def create_issue_comment(self, full_name: str, issue_number: int, body: str):
        # https://docs.github.com/en/rest/reference/issues#create-an-issue-comment
        response = requests.post(
            f"https://api.github.com/repos/{full_name}/issues/{issue_number}/comments",
            auth=self.auth,
            headers={"Accept": "application/vnd.github.v3+json"},
            json={"body": body},
        )
        response.raise_for_status()



import os.path
import subprocess
import instarepo.git
from instarepo.fixers.base import MissingFileFix


class MustHaveCliffTomlFix(MissingFileFix):
    """Ensures the configuration for git-cliff (cliff.toml) exists"""

    order = -100

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        super().__init__(git, "cliff.toml")

    def get_contents(self):
        template = os.path.join(os.path.dirname(__file__), "..", "..", "cliff.toml")
        with open(template, "r", encoding="utf-8") as file:
            return file.read()


class GenerateChangelogFix:
    """Generates changelog with git-cliff"""

    order = 100

    def __init__(self, git: instarepo.git.GitWorkingDir, **kwargs):
        self.git = git

    def run(self):
        if not self.git.isfile("cliff.toml"):
            return []
        changelog_file = "CHANGELOG.md"
        # load existing CHANGELOG contents, if it exists
        old_contents = ""
        if self.git.isfile(changelog_file):
            with open(self.git.join(changelog_file), "r", encoding="utf-8") as file:
                old_contents = file.read()
        # regenerate it
        subprocess.run(
            ["git-cliff", "-o", changelog_file], check=True, cwd=self.git.dir
        )
        # read the new CHANGELOG contents
        with open(self.git.join(changelog_file), "r", encoding="utf-8") as file:
            new_contents = file.read()
        # if same, no commit needed
        if old_contents == new_contents:
            return []
        self.git.add(changelog_file)
        msg = "chore(changelog): Updated changelog"
        self.git.commit(msg)
        return [msg]



import instarepo.github
import instarepo.repo_source


class ListCommand:
    def __init__(self, args):
        self.repo_source = (
            instarepo.repo_source.RepoSourceBuilder().with_args(args).build()
        )
        # remember sort order so that if we're sorting by a date field,
        # we'll be able to show that field on the last column
        self.sort = args.sort

    def run(self):
        repos = list(self.repo_source.get())
        if not repos:
            print("No repos found")
            return
        default_language = "N/A"
        max_repo_name_length = max(len(repo.name) for repo in repos)
        max_language_length = max(
            len(repo.language or default_language) for repo in repos
        )
        print(
            "{0:{1}s} {2:{3}s} {4}".format(
                "repo",
                max_repo_name_length,
                "language",
                max_language_length,
                date_column_title(self.sort),
            )
        )
        for repo in repos:
            print(
                "{0:{1}s} {2:{3}s} {4}".format(
                    repo.name,
                    max_repo_name_length,
                    repo.language or default_language,
                    max_language_length,
                    date_column_value(self.sort, repo),
                )
            )


def date_column_title(sort):
    if sort == "created":
        return "created at"
    elif sort == "updated":
        return "updated at"
    else:
        return "pushed at"


def date_column_value(sort, repo):
    if sort == "created":
        return repo.created_at
    elif sort == "updated":
        return repo.updated_at
    else:
        return repo.pushed_at



"""
Unit tests for main.py
"""

import pytest

from .main import create_command, parse_args
from .commands.analyze import AnalyzeCommand
from .commands.fix import FixCommand
from .commands.list import ListCommand


def test_parse_args_empty():
    """Empty arguments should throw error"""
    with pytest.raises(SystemExit):
        parse_args(args=[])


def test_parse_args_list_minimal():
    """Minimal list command parse"""
    result = parse_args(args=["list", "-u", "jdoe", "-t", "secret"])
    assert result.username == "jdoe"
    assert result.token == "secret"
    assert result.subparser_name == "list"


def test_parse_args_list_archived():
    """Tests the archived flag"""
    result = parse_args(args=["list", "-u", "jdoe", "-t", "secret"])
    assert result.archived == "deny"
    for valid_option in ["allow", "deny", "only"]:
        result = parse_args(
            args=["list", "-u", "jdoe", "-t", "secret", "--archived", valid_option]
        )
        assert result.archived == valid_option
    with pytest.raises(SystemExit):
        parse_args(args=["list", "-u", "jdoe", "-t", "secret", "--archived", "oops"])


def test_parse_args_fix_minimal():
    """Minimal fix command parse"""
    result = parse_args(args=["fix", "-u", "x", "--token", "sesame"])
    assert result.username == "x"
    assert result.token == "sesame"
    assert result.subparser_name == "fix"


def test_create_command_list():
    """Create ListCommand"""
    args = parse_args(args=["list", "-u", "jdoe", "-t", "secret"])
    cmd = create_command(args)
    assert isinstance(cmd, ListCommand)


def test_create_command_fix():
    """Create FixCommand"""
    args = parse_args(args=["fix", "-u", "jdoe", "-t", "secret"])
    cmd = create_command(args)
    assert isinstance(cmd, FixCommand)


def test_create_command_analyze():
    """Create AnalyzeCommand"""
    args = parse_args(
        args=["analyze", "-u", "jdoe", "-t", "secret", "--since", "2021-11-06"]
    )
    cmd = create_command(args)
    assert isinstance(cmd, AnalyzeCommand)


def test_parse_args_fix_fixer_selection():
    """Parse specific fixers"""
    args = parse_args(
        args=["fix", "-u", "jdoe", "-t", "secret", "--only-fixers", "dotnet", "maven"]
    )
    assert args.only_fixers == ["dotnet", "maven"]
    args = parse_args(
        args=["fix", "-u", "jdoe", "-t", "secret", "--except-fixers", "dotnet", "vb6"]
    )
    assert args.except_fixers == ["dotnet", "vb6"]
    with pytest.raises(SystemExit):
        parse_args(
            args=[
                "fix",
                "-u",
                "jdoe",
                "-t",
                "secret",
                "--except-fixers",
                "foo",
                "--skip-fixers",
                "bar",
            ]
        )



"""Unit tests for fix.py"""
import os

import pytest

import instarepo.fixers.base
import instarepo.fixers.changelog
import instarepo.fixers.dotnet
import instarepo.fixers.maven
import instarepo.git

from .fix import (
    all_fixer_classes,
    classes_in_module,
    create_composite_fixer,
    fixer_class_to_fixer_key,
    format_body,
    select_fixer_classes,
    try_get_fixer_order,
)


class TestFormatBody:
    """Unit tests for format_body"""

    def test_one_change(self):
        changes = ["Simple change"]
        expected_body = """The following fixes have been applied:
- Simple change
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body

    def test_two_changes(self):
        changes = ["Simple change", "Second change"]
        expected_body = """The following fixes have been applied:
- Simple change
- Second change
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body

    def test_convert_multi_line_to_indentation(self):
        changes = [
            """Complex change
Updated parent to 1.0
"""
        ]
        expected_body = """The following fixes have been applied:
- Complex change
  Updated parent to 1.0
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body


@pytest.fixture(params=[clz for clz in all_fixer_classes()])
def fixer_class(request):
    return request.param


def test_fixer_has_doc_string(fixer_class):  # pylint: disable=redefined-outer-name
    """Tests all fixer classes have a doc string"""
    assert fixer_class
    assert fixer_class.__doc__


def test_fixer_class_to_fixer_key():
    """Tests various fixer classes can be mapped to a key"""
    assert (
        fixer_class_to_fixer_key(instarepo.fixers.dotnet.MustHaveGitHubActionFix)
        == "dotnet.must_have_git_hub_action"
    )


def test_can_create_fixer(fixer_class):  # pylint: disable=redefined-outer-name
    """Tests that it is possible to instantiate all fixers"""
    mock_git = instarepo.git.GitWorkingDir("/tmp")
    mock_github = ()
    mock_repo = ()
    instance = fixer_class(
        git=mock_git, repo=mock_repo, github=mock_github, verbose=False
    )
    assert instance


def test_can_create_fixer_for_local_dir(
    fixer_class,
):  # pylint: disable=redefined-outer-name
    """Tests that it is possible to instantiate all fixers without repo/github instances"""
    mock_git = instarepo.git.GitWorkingDir("/tmp")
    instance = fixer_class(git=mock_git, repo=None, github=None, verbose=False)
    assert instance


class TestSelectFixerClasses:
    def test_returns_all_when_unfiltered(self):
        assert len(list(all_fixer_classes())) == len(select_fixer_classes())

    def test_filter_by_name(self):
        assert [
            instarepo.fixers.dotnet.DotNetFrameworkVersionFix,
            instarepo.fixers.dotnet.MustHaveGitHubActionFix,
        ] == select_fixer_classes(only_fixers=["dotnet"])

    def test_filter_by_two_names(self):
        assert [
            instarepo.fixers.dotnet.DotNetFrameworkVersionFix,
            instarepo.fixers.dotnet.MustHaveGitHubActionFix,
            instarepo.fixers.maven.MustHaveMavenGitHubWorkflowFix,
            instarepo.fixers.maven.MavenBadgesFix,
            instarepo.fixers.maven.UrlFix,
        ] == select_fixer_classes(only_fixers=["dotnet", "maven"])

    def test_cannot_use_only_and_except_together(self):
        with pytest.raises(ValueError):
            select_fixer_classes(only_fixers=["a"], except_fixers=["b"])

    def test_filter_except(self):
        assert [
            instarepo.fixers.dotnet.DotNetFrameworkVersionFix,
            instarepo.fixers.dotnet.MustHaveGitHubActionFix,
        ] == select_fixer_classes(
            except_fixers=[
                "changelog",
                "ci",
                "license",
                "maven",
                "missing_files",
                "pascal",
                "r",
                "vb",
            ]
        )

    def test_sort(self):
        assert [
            instarepo.fixers.changelog.MustHaveCliffTomlFix,
            instarepo.fixers.dotnet.DotNetFrameworkVersionFix,
            instarepo.fixers.dotnet.MustHaveGitHubActionFix,
            instarepo.fixers.changelog.GenerateChangelogFix,
        ] == select_fixer_classes(only_fixers=["dotnet", "changelog"])


def test_classes_in_module():
    assert [instarepo.git.GitWorkingDir] == list(classes_in_module(instarepo.git))


def test_try_get_fix_order():
    assert try_get_fixer_order(instarepo.fixers.dotnet.MustHaveGitHubActionFix) == 0
    assert try_get_fixer_order(instarepo.fixers.changelog.GenerateChangelogFix) == 100


def test_create_composite_fixer():
    # arrange
    git = instarepo.git.GitWorkingDir("/tmp")
    fixer_classes = [
        instarepo.fixers.changelog.MustHaveCliffTomlFix,
        instarepo.fixers.dotnet.DotNetFrameworkVersionFix,
    ]
    # act
    composite_fixer = create_composite_fixer(fixer_classes, git)
    # assert
    assert composite_fixer
    assert isinstance(composite_fixer, instarepo.fixers.base.CompositeFix)
    assert composite_fixer.rules
    assert isinstance(
        composite_fixer.rules[0], instarepo.fixers.changelog.MustHaveCliffTomlFix
    )
    assert isinstance(
        composite_fixer.rules[1], instarepo.fixers.dotnet.DotNetFrameworkVersionFix
    )



import functools
import logging
import os
import os.path
import subprocess
import instarepo.git

JCF_EXE = "C:\\opt\\jcf_243_exe\\JCF.exe"


def trim_trailing_whitespace(filename: str):
    with open(filename, "r", encoding="utf-8") as f:
        lines = f.readlines()
    lines = [line.rstrip() + "\n" for line in lines]
    with open(filename, "w", encoding="utf-8") as f:
        f.writelines(lines)


@functools.lru_cache(maxsize=None)
def find_jedi_cfg():
    local_app_data_dir = os.getenv("LOCALAPPDATA")
    if not local_app_data_dir:
        return ""
    if not os.path.isdir(local_app_data_dir):
        return ""
    lazarus_dir = os.path.join(local_app_data_dir, "lazarus")
    if not os.path.isdir(lazarus_dir):
        return ""
    cfg_file = os.path.join(lazarus_dir, "jcfsettings.cfg")
    if not os.path.isfile(cfg_file):
        return ""
    return cfg_file


def is_pascal_entry(entry):
    return entry.is_file() and (
        entry.name.endswith(".pas") or entry.name.endswith(".lpr")
    )


class AutoFormatFix:
    """Automatically formats Pascal files with JEDI code format"""

    def __init__(self, git: instarepo.git.GitWorkingDir, verbose: bool, **kwargs):
        self.git = git
        self.files = []
        self._fallback_ptop_cfg = None
        self.verbose = verbose

    def run(self):
        if not os.path.isfile(JCF_EXE):
            logging.debug("JEDI Code Format exe %s not found", JCF_EXE)
            return []
        if not find_jedi_cfg():
            logging.debug("JEDI Code Format cfg not found")
            return []
        with os.scandir(self.git.dir) as it:
            for entry in it:
                if is_pascal_entry(entry):
                    self._process(entry.path)
        if len(self.files) <= 0:
            return []
        msg = "chore: Auto-formatted Pascal files: " + ", ".join(self.files)
        self.git.commit(msg)
        return [msg]

    def _process(self, pas_file: str):
        with open(pas_file, "r", encoding="utf-8") as f:
            old_contents = f.read()
        rel_path = os.path.relpath(pas_file, self.git.dir)
        # pass through jcf
        if self.verbose:
            subprocess.run(self._build_args(rel_path), check=True, cwd=self.git.dir)
        else:
            subprocess.run(
                self._build_args(rel_path),
                check=True,
                cwd=self.git.dir,
                stdout=subprocess.PIPE,
            )
        # trim trailing whitespace
        trim_trailing_whitespace(pas_file)
        # check if we have changes
        with open(pas_file, "r", encoding="utf-8") as f:
            new_contents = f.read()
        if old_contents != new_contents:
            self.git.add(rel_path)
            self.files.append(rel_path)

    def _build_args(self, rel_pas_file: str):
        args = [
            JCF_EXE,
            "-config=" + find_jedi_cfg(),
            "-clarify",
            "-inplace",
            "-y",
            "-f",
            rel_pas_file,
        ]
        return args
